@@ -32,12 +32,20 @@ import re
 import statvfs
 import shutil
 import glob
+import time
 from DistUpgradeConfigParser import DistUpgradeConfig
 
-from aptsources import SourcesList, SourceEntry, Distribution, is_mirror
+# FIXME: we need this only for the later "isinstance()" check
+#        this should probably be solved in some different way
+from DistUpgradeViewText import DistUpgradeViewText
+
+from sourceslist import SourcesList, SourceEntry, is_mirror
+from distro import Distribution, get_distro
+
 from gettext import gettext as _
 import gettext
 from DistUpgradeCache import MyCache
+from DistUpgradeApport import *
 
 class AptCdrom(object):
     def __init__(self, view, path):
@@ -60,7 +68,8 @@ class AptCdrom(object):
         if backup_ext:
             cdromstate = os.path.join(apt_pkg.Config.FindDir("Dir::State"),
                                       apt_pkg.Config.Find("Dir::State::cdroms"))
-            shutil.copy(cdromstate, cdromstate+backup_ext)
+            if os.path.exists(cdromstate):
+                shutil.copy(cdromstate, cdromstate+backup_ext)
         # do the actual work
         apt_pkg.Config.Set("Acquire::cdrom::mount",self.cdrompath)
         apt_pkg.Config.Set("APT::CDROM::NoMount","true")
@@ -116,6 +125,11 @@ class DistUpgradeControler(object):
         else:
             cdrompath = None
         self.aptcdrom = AptCdrom(distUpgradeView, cdrompath)
+
+        # we act differently in server mode
+        self.serverMode = False
+        if self.options and self.options.mode == "server":
+            self.serverMode = True
         
         # the configuration
         self.config = DistUpgradeConfig(datadir)
@@ -133,15 +147,122 @@ class DistUpgradeControler(object):
         apt_pkg.Config.Set("Debug::pkgProblemResolver","true")
         apt_pkg.Config.Set("Debug::pkgDepCache::AutoInstall","true")
         fd = os.open("/var/log/dist-upgrade/apt.log",
-                     os.O_RDWR|os.O_CREAT|os.O_APPEND, 0644)
-        os.dup2(fd,1)
-        os.dup2(fd,2)
+                     os.O_RDWR|os.O_CREAT|os.O_APPEND|os.O_SYNC, 0644)
+        # log the complete output if we do not run in text-mode
+        if not isinstance(self._view, DistUpgradeViewText):
+            os.dup2(fd,2)
+            os.dup2(fd,1)
+        self.logfd = fd
 
     def openCache(self):
-        self.cache = MyCache(self.config, self._view.getOpCacheProgress())
+        self.cache = MyCache(self.config,
+                             self._view,
+                             self._view.getOpCacheProgress())
 
+    def _sshMagic(self):
+        """ this will check for server mode and if we run over ssh.
+            if this is the case, we will ask and spawn a additional
+            daemon (to be sure we have a spare one around in case
+            of trouble)
+        """
+        if (self.serverMode and
+            (os.environ.has_key("SSH_CONNECTION") or
+             os.environ.has_key("SSH_TTY"))):
+            port = 9004
+            res = self._view.askYesNoQuestion(
+                _("Continue running under SSH?"),
+                _("This session appears to be running under ssh. "
+                  "It is not recommended to perform a upgrade "
+                  "over ssh currently because in case of failure "
+                "it is harder to recover.\n\n"
+                  "If you continue, a additional ssh daemon will be "
+                  "started at port '%s'.\n"
+                  "Do you want to continue?") % port)
+            # abort
+            if res == False:
+                sys.exit(1)
+            res = subprocess.call(["/usr/sbin/sshd","-p",str(port)])
+            if res == 0:
+                self._view.information(
+                    _("Starting additional sshd"),
+                    _("To make recovery in case of failure easier a "
+                      "additional sshd will be started on port '%s'. "
+                      "If anything goes wrong with the running ssh "
+                      "you can still connect to the additional one.\n"
+                      ) % port)
+
+    def _tryUpdateSelf(self):
+        """ this is a helper that is run if we are started from a CD
+            and we have network - we will then try to fetch a update
+            of ourself
+        """  
+        from MetaRelease import MetaReleaseCore
+        from DistUpgradeFetcherSelf import DistUpgradeFetcherSelf
+        # FIXME: during testing, we want "useDevelopmentRelease"
+        #        but not after the release
+        m = MetaReleaseCore(useDevelopmentRelease=False)
+        # this will timeout eventually
+        while m.downloading:
+            self._view.processEvents()
+            time.sleep(0.1)
+        if m.new_dist is None:
+            logging.error("No new dist found")
+            return False
+        # we have a new dist
+        progress = self._view.getFetchProgress()
+        fetcher = DistUpgradeFetcherSelf(new_dist=m.new_dist,
+                                         progress=progress,
+                                         options=self.options,
+                                         view=self._view)
+        fetcher.run()
+
+    def _pythonSymlinkCheck(self):
+        """ check if sanity check, that /usr/bin/python points to the default
+            python version. Users tend to modify this symlink, which then
+            breaks stuff in obscure ways (Ubuntu #75557)
+        """
+        logging.debug("_pythonSymlinkCheck run")
+        from ConfigParser import SafeConfigParser, NoOptionError
+        if os.path.exists('/usr/share/python/debian_defaults'):
+            config = SafeConfigParser()
+            config.readfp(file('/usr/share/python/debian_defaults'))
+            try:
+                expected_default = config.get('DEFAULT', 'default-version')
+            except NoOptionError:
+                logging.debug("no default version for python found in '%s'" % config)
+                return False
+            try:
+                fs_default_version = os.readlink('/usr/bin/python')
+            except OSError, e:
+                logging.error("os.readlink failed (%s)" % e)
+                return False
+            if not fs_default_version in (expected_default, os.path.join('/usr/bin', expected_default)):
+                logging.debug("python symlink points to: '%s', but expected is '%s' or '%s'" % (fs_default_version, expected_default, os.path.join('/usr/bin', expected_default)))
+                return False
+        return True
+    
     def prepare(self):
         """ initial cache opening, sanity checking, network checking """
+        # first check if that is a good upgrade
+        release = subprocess.Popen(["lsb_release","-c","-s"],
+                                   stdout=subprocess.PIPE).communicate()[0].strip()
+        logging.debug("lsb-release: '%s'" % release)
+        if not (release == self.fromDist or release == self.toDist):
+            logging.error("Bad upgrade: '%s' != '%s' " % (release, self.fromDist))
+            self._view.error(_("Can not upgrade"),
+                             _("A upgrade from '%s' to '%s' is not "
+                               "supoprted with this tool." % (release, self.toDist)))
+            sys.exit(1)
+        # do the ssh check and warn if we run under ssh
+        self._sshMagic()
+        # check python version
+        if not self._pythonSymlinkCheck():
+            logging.error("pythonSymlinkCheck() failed, aborting")
+            self._view.error(_("Can not upgrade"),
+                             _("Your python install is corrupted. "
+                               "Please fix the '/usr/bin/python' symlink."))
+            sys.exit(1)
+        # open cache
         try:
             self.openCache()
         except SystemError, e:
@@ -152,38 +273,57 @@ class DistUpgradeControler(object):
         # FIXME: we may try to find out a bit more about the network
         # connection here and ask more  inteligent questions
         if self.aptcdrom and self.options and self.options.withNetwork == None:
-            res = self._view.askYesNoQuestion(_("Fetch data from the network for the upgrade?"),
-                                              _("The upgrade can use the network to check "
-                                                "the latest updates and to fetch packages that are not on the "
-                                                "current CD.\n"
-                                                "If you have fast or inexpensive network access you should answer "
-                                                "'Yes' here. If networking is expensive for you choose 'No'.")
+            res = self._view.askYesNoQuestion(_("Include latest updates from the Internet?"),
+                                              _("The upgrade process can automatically download "
+                                                "the latest updates and install them during the "
+                                                "upgrade.  The upgrade will take longer, but when "
+                                                "it is complete, your system will be fully up to "
+                                                "date.  You can choose not to do this, but you "
+                                                "should install the latest updates soon after "
+                                                "upgrading."),
+                                              'Yes'
                                               )
             self.useNetwork = res
             logging.debug("useNetwork: '%s' (selected by user)" % res)
+            if res:
+                self._tryUpdateSelf()
         return True
 
     def rewriteSourcesList(self, mirror_check=True):
         logging.debug("rewriteSourcesList()")
 
-        # enable main (we always need this!)
-        distro = Distribution()
-        distro.get_sources(self.sources)
-        # make sure that main is enabled
-        distro.enable_component(self.sources, "main")
-
+        # check if we need to enable main
+        if mirror_check == True and self.useNetwork:
+            # now check if the base-meta pkgs are available in
+            # the archive or only available as "now"
+            # -> if not that means that "main" is missing and we
+            #    need to  enable it
+            for pkgname in self.config.getlist("Distro","BaseMetaPkgs"):
+                if ((len(self.cache[pkgname].candidateOrigin) == 0)
+                    or
+                    (len(self.cache[pkgname].candidateOrigin) == 1 and
+                     self.cache[pkgname].candidateOrigin[0].archive == "now")):
+                    distro = get_distro()
+                    distro.get_sources(self.sources)
+                    distro.enable_component("main")
+                    break
+            
         # this must map, i.e. second in "from" must be the second in "to"
         # (but they can be different, so in theory we could exchange
         #  component names here)
         fromDists = [self.fromDist,
                      self.fromDist+"-security",
                      self.fromDist+"-updates",
-                     self.fromDist+"-backports"
+                     self.fromDist+"-proposed",
+                     self.fromDist+"-backports",
+                     self.fromDist+"-commercial"
                     ]
         toDists = [self.toDist,
                    self.toDist+"-security",
                    self.toDist+"-updates",
-                   self.toDist+"-backports"
+                   self.toDist+"-proposed",
+                   self.toDist+"-backports",
+                   self.toDist+"-commercial"
                    ]
 
         # list of valid mirrors that we can add
@@ -201,28 +341,42 @@ class DistUpgradeControler(object):
             
             # we disable breezy cdrom sources to make sure that demoted
             # packages are removed
-            if entry.uri.startswith("cdrom:") and entry.dist == "breezy":
+            if entry.uri.startswith("cdrom:") and entry.dist == self.fromDist:
                 entry.disabled = True
                 continue
             # ignore cdrom sources otherwise
             elif entry.uri.startswith("cdrom:"):
                 continue
-                
+
+            # special case for old-releases.ubuntu.com, auto transition
+            # them back to archive.ubuntu.com - now this is a problem
+            # of course for people upgrading from EOL release to a 
+            # EOL release
+            if (not entry.disabled and
+                entry.uri.startswith("http://old-releases.ubuntu.com/ubuntu")):
+                entry.uri = "http://archive.ubuntu.com/ubuntu"
+                logging.debug("transitioning old-releases.ubuntu.com to '%s' " % entry)
+                continue
+
             logging.debug("examining: '%s'" % entry)
             # check if it's a mirror (or offical site)
             validMirror = False
             for mirror in valid_mirrors:
                 if not mirror_check or is_mirror(mirror,entry.uri):
                     validMirror = True
-                    # security is a special case
-                    res = not entry.uri.startswith("http://security.ubuntu.com") and not entry.disabled
+                    # disabled/security/commercial are special cases
+                    validTo = True
+                    if (entry.disabled or
+                        entry.uri.startswith("http://security.ubuntu.com") or
+                        entry.uri.startswith("http://archive.canonical.com")):
+                        validTo = False
                     if entry.dist in toDists:
                         # so the self.sources.list is already set to the new
                         # distro
                         logging.debug("entry '%s' is already set to new dist" % entry)
-                        foundToDist |= res
+                        foundToDist |= validTo
                     elif entry.dist in fromDists:
-                        foundToDist |= res
+                        foundToDist |= validTo
                         entry.dist = toDists[fromDists.index(entry.dist)]
                         logging.debug("entry '%s' updated to new dist" % entry)
                     else:
@@ -246,10 +400,10 @@ class DistUpgradeControler(object):
         if not self.rewriteSourcesList(mirror_check=True):
             logging.error("No valid mirror found")
             res = self._view.askYesNoQuestion(_("No valid mirror found"),
-                             _("While scaning your repository "
+                             _("While scanning your repository "
                                "information no mirror entry for "
                                "the upgrade was found."
-                               "This cam happen if you run a internal "
+                               "This can happen if you run a internal "
                                "mirror or if the mirror information is "
                                "out of date.\n\n"
                                "Do you want to rewrite your "
@@ -309,7 +463,8 @@ class DistUpgradeControler(object):
                              _("Some third party entries in your sources.list "
                                "were disabled. You can re-enable them "
                                "after the upgrade with the "
-                               "'software-properties' tool or with synaptic."
+                               "'software-properties' tool or "
+                               "your package manager."
                                ))
         return True
 
@@ -319,19 +474,25 @@ class DistUpgradeControler(object):
         inst = []
         up = []
         rm = []
+        held = []
         for pkg in self.cache:
             if pkg.markedInstall: inst.append(pkg.name)
             elif pkg.markedUpgrade: up.append(pkg.name)
             elif pkg.markedDelete: rm.append(pkg.name)
+            elif (pkg.isInstalled and pkg.isUpgradable): held.append(pkg.name)
+        logging.debug("Held-back: %s" % " ".join(held))
         logging.debug("Remove: %s" % " ".join(rm))
         logging.debug("Install: %s" % " ".join(inst))
         logging.debug("Upgrade: %s" % " ".join(up))
+        
 
     def doPreUpgrade(self):
         # FIXME: check out what packages are downloadable etc to
         # compare the list after the update again
         self.obsolete_pkgs = self.cache._getObsoletesPkgs()
         self.foreign_pkgs = self.cache._getForeignPkgs(self.origin, self.fromDist, self.toDist)
+        if self.serverMode:
+            self.tasks = self.cache.installedTasks
         logging.debug("Foreign: %s" % " ".join(self.foreign_pkgs))
         logging.debug("Obsolete: %s" % " ".join(self.obsolete_pkgs))
 
@@ -344,7 +505,7 @@ class DistUpgradeControler(object):
         # FIXME: retry here too? just like the DoDistUpgrade?
         #        also remove all files from the lists partial dir!
         currentRetry = 0
-        maxRetries = int(self.config.get("Network","MaxRetries"))
+        maxRetries = self.config.getint("Network","MaxRetries")
         while currentRetry < maxRetries:
             try:
                 res = self.cache.update(progress)
@@ -354,7 +515,8 @@ class DistUpgradeControler(object):
                 continue
             # no exception, so all was fine, we are done
             return True
-                
+
+        logging.error("doUpdate() failed complettely")
         self._view.error(_("Error during update"),
                          _("A problem occured during the update. "
                            "This is usually some sort of network "
@@ -372,64 +534,79 @@ class DistUpgradeControler(object):
                     "packages of former installations using "
                     "'sudo apt-get clean'.")
 
-        # gather/log some staticts
+        class FreeSpace(object):
+            " helper class that represents the free space on each mounted fs "
+            def __init__(self, initialFree):
+                self.free = initialFree
+
+        def make_fs_id(d):
+            """ return 'id' of a directory so that directories on the
+                same filesystem get the same id (simply the mount_point)
+            """
+            for mount_point in mounted:
+                if d.startswith(mount_point):
+                    return mount_point
+            return "/"
+
+        # this is all a bit complicated
+        # 1) check what is mounted (in mounted)
+        # 2) create FreeSpace objects for the dirs we are interessted in
+        #    (mnt_map)
+        # 3) use the  mnt_map to check if we have enough free space and
+        #    if not tell the user how much is missing
+        mounted = []
         mnt_map = {}
-        for d in ["/","/usr","/var","/boot"]:
+        fs_free = {}
+        for line in open("/proc/mounts"):
+            (what, where, fs, options, a, b) = line.split()
+            if not where in mounted:
+                mounted.append(where)
+        # make sure mounted is sorted by longest path
+        mounted.sort(cmp=lambda a,b: cmp(len(a),len(b)), reverse=True)
+        archivedir = apt_pkg.Config.FindDir("Dir::Cache::archives")
+        for d in ["/","/usr","/var","/boot", archivedir, "/home"]:
+            fs_id = make_fs_id(d)
             st = os.statvfs(d)
             free = st[statvfs.F_BAVAIL]*st[statvfs.F_FRSIZE]
-            if st in mnt_map:
-                logging.debug("Dir %s mounted on %s" % (d,mnt_map[st]))
+            if fs_id in mnt_map:
+                logging.debug("Dir %s mounted on %s" % (d,mnt_map[fs_id]))
+                fs_free[d] = fs_free[mnt_map[fs_id]]
             else:
                 logging.debug("Free space on %s: %s" % (d,free))
-                mnt_map[st] = d
+                mnt_map[fs_id] = d
+                fs_free[d] = FreeSpace(free)
         del mnt_map
+        logging.debug("fs_free contains: '%s'" % fs_free)
 
-        # first check for /var (or where the archives are downloaded too)
-        archivedir = apt_pkg.Config.FindDir("Dir::Cache::archives")
-        st_archivedir = os.statvfs(archivedir)
-        free = st_archivedir[statvfs.F_BAVAIL]*st_archivedir[statvfs.F_FRSIZE]
-        logging.debug("required download: %s " % self.cache.requiredDownload)
-        logging.debug("free on %s: %s " % (archivedir, free))
-        if self.cache.requiredDownload > free:
-            free_at_least = apt_pkg.SizeToStr(self.cache.requiredDownload-free)
-            self._view.error(err_sum, err_long % (free_at_least,archivedir))
-            return False
-        
-        # then check for /usr assuming that all the data goes into /usr
-        # this won't catch space problems when e.g. /boot,/usr/,/ are all
-        # seperated partitions, but with a fragmented
-        # patition layout we can't do a lot better because we don't know
-        # the space-requirements on a per dir basis inside the deb without
-        # looking into each
-        logging.debug("need additional space: %s" % self.cache.additionalRequiredSpace)
-        dir = "/usr"
-        st_usr = os.statvfs(dir)
-        if st_archivedir == st_usr:
-            # we are on the same filesystem, so we need to take the space
-            # for downloading the debs into account
-            free -= self.cache.requiredDownload
-            logging.debug("/usr on same fs as %s, taking dl-size into account, new free: %s" % (archivedir, free))
-        else:
-            free = st_usr[statvfs.F_BAVAIL]*st_usr[statvfs.F_FRSIZE]
-            logging.debug("/usr on different fs than %s, free: %s" % (archivedir, free))
+        # we check for various sizes:
+        # archivedir is were we download the debs
+        # /usr is assumed to get *all* of the install space (incorrect,
+        #      but as good as we can do currently + savety buffer
+        # /boot is assumed to get at least 50 Mb
+        # /     has a small savety buffer as well
+        for (dir, size) in [(archivedir, self.cache.requiredDownload),
+                            ("/usr", self.cache.additionalRequiredSpace),
+                            ("/usr", 50*1024*1024),  # savetfy buffer /usr
+                            ("/boot", 40*1024*1024), # savetfy buffer /boot
+                            ("/", 10*1024*1024),     # small savetfy buffer /
+                           ]:
+            logging.debug("dir '%s' needs '%s' of '%s' (%f)" % (dir, size, fs_free[dir], fs_free[dir].free))
+            fs_free[dir].free -= size
+            if fs_free[dir].free < 0:
+                free_at_least = apt_pkg.SizeToStr(float(abs(fs_free[dir].free)+1))
+                logging.error("not enough free space on %s (missing %s)" % (dir, free_at_least))
+                self._view.error(err_sum, err_long % (free_at_least,dir))
+                return False
 
-        safety_buffer = 1024*1024*100 # 100 Mb
-        logging.debug("using safety buffer: %s" % safety_buffer)
-        if (self.cache.additionalRequiredSpace+safety_buffer) > free:
-            free_at_least = apt_pkg.SizeToStr(self.cache.additionalRequiredSpace+safety_buffer-free)
-            logging.error("not enough free space, we need addional %s" % free_at_least)
-            self._view.error(err_sum, err_long % (free_at_least,dir))
-            return False
-
-        # FIXME: we should try to esitmate if "/" has enough free space,
-        # linux-restricted-modules and linux-image- are both putting there
-        # modules there and those take a lot of space
             
         return True
 
     def askDistUpgrade(self):
-        if not self.cache.distUpgrade(self._view):
+        if not self.cache.distUpgrade(self._view, self.serverMode, self.logfd):
             return False
+        if self.serverMode:
+            if not self.cache.installTasks(self.tasks):
+                return False
         changes = self.cache.getChanges()
         # log the changes for debuging
         self._logChanges()
@@ -450,21 +627,28 @@ class DistUpgradeControler(object):
         fprogress = self._view.getFetchProgress()
         iprogress = self._view.getInstallProgress(self.cache)
         # retry the fetching in case of errors
-        maxRetries = int(self.config.get("Network","MaxRetries"))
+        maxRetries = self.config.getint("Network","MaxRetries")
         while currentRetry < maxRetries:
             try:
                 res = self.cache.commit(fprogress,iprogress)
             except SystemError, e:
+                logging.error("SystemError from cache.commit(): %s" % e)
+                # check if the installprogress catched a pkgfailure, if not, generate a fallback here
+                if iprogress.pkg_failures == 0:
+                    errormsg = "SystemError in cache.commit(): %s" % e
+                    apport_pkgfailure("update-manager", errormsg)
+                # invoke the frontend now
+                msg = _("The upgrade aborts now. Your system "
+                        "could be in an unusable state. A recovery "
+                        "will run now (dpkg --configure -a).")
+                if not run_apport():
+                    msg += _("\n\nPlease report this bug against the 'update-manager' "
+                             "package and include the files in /var/log/dist-upgrade/ "
+                             "in the bugreport.\n"
+                             "%s" % e)
+                self._view.error(_("Could not install the upgrades"), msg)
                 # installing the packages failed, can't be retried
                 self._view.getTerminal().call(["dpkg","--configure","-a"])
-                self._view.error(_("Could not install the upgrades"),
-                                 _("The upgrade aborts now. Your system "
-                                   "could be in an unusable state. A recovery "
-                                   "was run (dpkg --configure -a).\n\n"
-                                   "Please report this bug against the 'update-manager' "
-                                   "package and include the files in /var/log/dist-upgrade/ "
-                                   "in the bugreport."),
-                                 "%s" % e)
                 return False
             except IOError, e:
                 # fetch failed, will be retried
@@ -475,7 +659,7 @@ class DistUpgradeControler(object):
             return True
         
         # maximum fetch-retries reached without a successful commit
-        logging.debug("giving up on fetching after maximum retries")
+        logging.error("giving up on fetching after maximum retries")
         self._view.error(_("Could not download the upgrades"),
                          _("The upgrade aborts now. Please check your "\
                            "internet connection or "\
@@ -494,6 +678,10 @@ class DistUpgradeControler(object):
         now_foreign = self.cache._getForeignPkgs(self.origin, self.fromDist, self.toDist)
         logging.debug("Obsolete: %s" % " ".join(now_obsolete))
         logging.debug("Foreign: %s" % " ".join(now_foreign))
+        # check if we actually want obsolete removal
+        if not self.config.getWithDefault("Distro","RemoveObsoletes", True):
+            logging.debug("Skipping obsolete Removal")
+            return True
 
         # now get the meta-pkg specific obsoletes and purges
         for pkg in self.config.getlist("Distro","MetaPkgs"):
@@ -502,8 +690,8 @@ class DistUpgradeControler(object):
         logging.debug("forced_obsoletes: %s", self.forced_obsoletes)
 
         # check what packages got demoted
-        demotions_file = self.config.get("Distro","Demotions")
         demotions = set()
+        demotions_file = self.config.get("Distro","Demotions")
         if os.path.exists(demotions_file):
             map(lambda pkgname: demotions.add(pkgname.strip()),
                 filter(lambda line: not line.startswith("#"),
@@ -542,6 +730,7 @@ class DistUpgradeControler(object):
         logging.debug("Start checking for obsolete pkgs")
         for pkgname in remove_candidates:
             if pkgname not in self.foreign_pkgs:
+                self._view.processEvents()
                 if not self.cache._tryMarkObsoleteForRemoval(pkgname, remove_candidates, self.foreign_pkgs):
                     logging.debug("'%s' scheduled for remove but not in remove_candiates, skipping", pkgname)
         logging.debug("Finish checking for obsolete pkgs")
@@ -560,11 +749,99 @@ class DistUpgradeControler(object):
             try:
                 res = self.cache.commit(fprogress,iprogress)
             except (SystemError, IOError), e:
+                logging.error("cache.commit() in doPostUpgrade() failed: %s" % e)
                 self._view.error(_("Error during commit"),
-                                 _("Some problem occured during the clean-up. "
+                                 _("A problem occured during the clean-up. "
                                    "Please see the below message for more "
                                    "information. "),
                                    "%s" % e)
+        # now run the post-upgrade fixup scripts (if any)
+        for script in self.config.getlist("Distro","PostInstallScripts"):
+            logging.debug("Runing PostInstallScript: '%s'" % script)
+            try:
+                # work around kde being clever
+                if script.startswith("./"):
+                    os.chmod(script, 0755)
+                self._view.getTerminal().call([script], hidden=True)
+            except Exception, e:
+                logging.error("gor error from PostInstallScript %s (%s)" % (script, e))
+        # now run the quirksHandler 
+        quirksFuncName = "%sQuirks" % self.config.get("Sources","To")
+        func = getattr(self, quirksFuncName, None)
+        if func is not None:
+            func()
+
+    def _rewriteFstab(self):
+        " convert /dev/{hd?,scd0} to /dev/cdrom for the feisty upgrade "
+        logging.debug("_rewriteFstab()")
+        replaced = 0
+        lines = []
+        # we have one cdrom to convert
+        for line in open("/etc/fstab"):
+            line = line.strip()
+            if line == '' or line.startswith("#"):
+                lines.append(line)
+                continue
+            try:
+                (device, mount_point, fstype, options, a, b) = line.split()
+            except Exception, e:
+                logging.error("can't parse line '%s'" % line)
+                lines.append(line)
+                continue
+            # edgy kernel has /dev/cdrom -> /dev/hd?
+            # feisty kernel (for a lot of chipsets) /dev/cdrom -> /dev/scd0
+            # this breaks static mounting (LP#86424)
+            #
+            # we convert here to /dev/cdrom only if current /dev/cdrom
+            # points to the device in /etc/fstab already. this ensures
+            # that we don't break anything or that we get it wrong
+            # for systems with two (or more) cdroms. this is ok, because
+            # we convert under the old kernel
+            if ("iso9660" in fstype and
+                device != "/dev/cdrom" and
+                os.path.exists("/dev/cdrom") and
+                os.path.realpath("/dev/cdrom") == device
+                ):
+                logging.debug("replacing '%s' " % line)
+                line = line.replace(device,"/dev/cdrom")
+                logging.debug("replacied line is '%s' " % line)
+                replaced += 1
+            lines.append(line)
+        # we have converted a line (otherwise we would have exited already)
+        if replaced > 0:
+            logging.debug("writing new /etc/fstab")
+            shutil.copy("/etc/fstab","/etc/fstab.edgy")
+            open("/etc/fstab","w").write("\n".join(lines))
+        return True
+
+    def _checkAdminGroup(self):
+        " check if the current sudo user is in the admin group "
+        logging.debug("_checkAdminGroup")
+        import grp
+        try:
+            admin_group = grp.getgrnam("admin").gr_mem
+        except KeyError, e:
+            logging.warning("System has no admin group (%s)" % e)
+            subprocess.call(["addgroup","--system","admin"])
+        admin_group = grp.getgrnam("admin").gr_mem
+        # if the current SUDO_USER is not in the admin group
+        # we add him - this is no security issue because
+        # the user is already root so adding him to the admin group
+        # does not change anything
+        if (os.environ.has_key("SUDO_USER") and
+            not os.environ["SUDO_USER"] in admin_group):
+            admin_user = os.environ["SUDO_USER"]
+            logging.info("SUDO_USER=%s is not in admin group" % admin_user)
+            cmd = ["usermod","-a","-G","admin",admin_user]
+            res = subprocess.call(cmd)
+            logging.debug("cmd: %s returned %i" % (cmd, res))
+        
+
+    def feistyQuirks(self):
+        """ this function works around quirks in the edgy->feisty upgrade """
+        logging.debug("running Controler.feistyQuirks handler")
+        self._rewriteFstab()
+        self._checkAdminGroup()
             
     def abort(self):
         """ abort the upgrade, cleanup (as much as possible) """
@@ -607,7 +884,6 @@ class DistUpgradeControler(object):
             pkg = self.cache[pkgname]
             # look for the right version (backport)
             for ver in pkg._pkg.VersionList:
-                print ver.VerStr
                 if self.config.get("Backports","VersionIdent") in ver.VerStr:
                     break
             else:
@@ -615,7 +891,6 @@ class DistUpgradeControler(object):
                 raise Exception, "No backport found!?!"
                 return False
             if ver.FileList == None:
-                print "No FileList for: %s " % self._pkg.Name()
                 return False
             f, index = ver.FileList.pop(0)
             pkg._records.Lookup((f,index))
@@ -662,12 +937,13 @@ class DistUpgradeControler(object):
         os.execve(sys.argv[0],args, os.environ)
 
     # this is the core
-    def edgyUpgrade(self):
+    def fullUpgrade(self):
         # sanity check (check for ubuntu-desktop, brokenCache etc)
         self._view.updateStatus(_("Checking package manager"))
         self._view.setStep(1)
         
         if not self.prepare():
+            logging.error("self.prepared() failed")
             self._view.error(_("Preparing the upgrade failed"),
                              _("Preparing the system for the upgrade "
                                "failed. Please report this as a bug "
@@ -734,7 +1010,10 @@ class DistUpgradeControler(object):
         if not self.askDistUpgrade():
             self.abort()
 
-        self._view.updateStatus(_("Upgrading"))            
+        # kill update-notifier now to supress reboot required
+        subprocess.call(["killall","-q","update-notifier"])
+        # no do the upgrade
+        self._view.updateStatus(_("Upgrading"))
         if not self.doDistUpgrade():
             # don't abort here, because it would restore the sources.list
             sys.exit(1) 
@@ -749,9 +1028,19 @@ class DistUpgradeControler(object):
         self._view.updateStatus(_("System upgrade is complete."))            
         # FIXME should we look into /var/run/reboot-required here?
         if self._view.confirmRestart():
-            subprocess.call(["reboot"])
+            p = subprocess.Popen("/sbin/reboot")
+            sys.exit(0)
         
     def run(self):
-        self.edgyUpgrade()
+        self.fullUpgrade()
 
 
+if __name__ == "__main__":
+    from DistUpgradeView import DistUpgradeView
+    from DistUpgradeCache import MyCache
+    v = DistUpgradeView()
+    dc = DistUpgradeControler(v)
+    #dc.openCache()
+    #dc._checkFreeSpace()
+    #dc._rewriteFstab()
+    dc._checkAdminGroup()
